/*
    dp : An implementation of the davis puttnam procedure for solving
         propositional satisfiability problems.
    Copyright (C) 1993  Peter Barth

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ------------------------------------------------------------------
   heuristics.c
   It should be fairly easy to link in a personal heuristic
*/

#include "datadefs.h"
#include "heuristics.h"
#include "stdlib.h"

/**************************************************************************/
/* The heuristics                                                         */
/**************************************************************************/

/**************************************************************************/
/* the simplest heuristic,namely none -> get the first active variable positiv */
/**************************************************************************/
int heuristic0(silist av,silist ac)
{
  return av->next->key;
}

/**************************************************************************
 * Jeroslow-Wang 
 * As in "Solving propositional satsifiability problems"
 **************************************************************************/

/* we don't need floating point operations if we approximate like this.
   Only the relations need to be the same
 */
int twopow(int y)
{
  if (y > 20 || y <= 0)
    return(1);
  return (1 << (20 - y));
}

int heuristic1(silist av,silist ac) 
{
  silist iav = av->next;
  silist iac;
  silist maxav = av;
  int* varinclauses;
  int posocc = 0;
  int negocc = 0;
  int maxocc = 0;
  int max_sign = 0;
  
  do {
    varinclauses = lit_to_clauses[iav->key];
    iac = ac->next;
    posocc = negocc = 0;
    for(;;) /* ugly endless loop for speed, direct abort if end condition */
      {
	if(GETtaclauseid(iac->key) < abs(*varinclauses))
	  {
	    iac = iac->next; if (iac->next == iac) break;
	  }
	else if(GETtaclauseid(iac->key) == abs(*varinclauses))
	  {
	    if (*varinclauses > 0) 
	      posocc += twopow(GETtaclauselen((int) iac->key)); 
	    else
	      negocc += twopow(GETtaclauselen((int) iac->key)); 
	    iac = iac->next; if (iac->next == iac)  break;
	    varinclauses++; if (*varinclauses == 0) break;
	  }  
	else      
	  { varinclauses++; if (*varinclauses == 0) break; }
      } /* while (*varinclauses && (iac->next != iac)) */
    if (posocc > maxocc) 
      { maxav = iav; maxocc = posocc; max_sign = 1; }
    if (negocc > maxocc) 
      { maxav = iav; maxocc = negocc; max_sign = -1; }
    iav = iav->next;
  }  while(iav->next != iav);
  return max_sign*maxav->key;
}


/*********************************************************** 
 * variable with most often occurence, with most often sign 
 **********************************************************/

int heuristic2(silist av,silist ac) 
{
  silist iav = av->next;
  silist iac;
  silist maxav = av;
  int* varinclauses;
  int no_pos_occs = 0;
  int no_neg_occs = 0;
  int max_no_occs = 0;
  int max_sign = 0;
  
  do {
    varinclauses = lit_to_clauses[iav->key];
    iac = ac->next;
    no_pos_occs = no_neg_occs = 0;
    for(;;) /* ugly endless loop for speed, direct abort if end condition */
      {
	if(GETtaclauseid(iac->key) < abs(*varinclauses))
	  {
	    iac = iac->next; if (iac->next == iac) break;
	  }
	else if(GETtaclauseid(iac->key) == abs(*varinclauses))
	  {
	    if (*varinclauses > 0) 
	      no_pos_occs++;	
	    else
	      no_neg_occs++;
	    iac = iac->next; if (iac->next == iac)  break;
	    varinclauses++; if (*varinclauses == 0) break;
	  }  
	else      
	  { varinclauses++; if (*varinclauses == 0) break; }
      } /* while (*varinclauses && (iac->next != iac)) */
    if ((no_pos_occs+no_neg_occs) > max_no_occs) 
      { 
	maxav = iav; 
	max_no_occs = no_pos_occs+no_neg_occs;
	max_sign = no_pos_occs > no_neg_occs ? 1 : -1;
      }
    iav = iav->next;
  }  while(iav->next != iav);
  return max_sign*maxav->key;
}


/*********************************************************** 
 * A variant of jw-heuristic
 * For selecting a variable we maximize the *sum* of the valuation
 * of -V and V and consider not only every literal on its own.
 * (Idea coming after email communication with Allen van Gelder May 1993)
 **********************************************************/

int heuristic3(silist av,silist ac) 
{
  silist iav = av->next;
  silist iac;
  silist maxav = av;
  int* varinclauses;
  int posocc = 0;
  int negocc = 0;
  int maxocc = 0;
  int max_sign = 0;
  
  do {
    varinclauses = lit_to_clauses[iav->key];
    iac = ac->next;
    posocc = negocc = 0;
    for(;;) /* ugly endless loop for speed, direct abort if end condition */
      {
	if(GETtaclauseid(iac->key) < abs(*varinclauses))
	  {
	    iac = iac->next; if (iac->next == iac) break;
	  }
	else if(GETtaclauseid(iac->key) == abs(*varinclauses))
	  {
	    if (*varinclauses > 0) 
	      posocc += twopow(GETtaclauselen((int) iac->key)); 
	    else
	      negocc += twopow(GETtaclauselen((int) iac->key)); 
	    iac = iac->next; if (iac->next == iac)  break;
	    varinclauses++; if (*varinclauses == 0) break;
	  }  
	else      
	  { varinclauses++; if (*varinclauses == 0) break; }
      } /* while (*varinclauses && (iac->next != iac)) */
    if (posocc + negocc > maxocc) 
      { 
	maxav = iav; 
	maxocc = negocc +  posocc; 
	max_sign = posocc > negocc ? 1 : -1; 
      }
    iav = iav->next;
  }  while(iav->next != iav);
  return max_sign*maxav->key;
}


/**************************************************************************
 * randomly choosen literal
 **************************************************************************/
int heuristic4(silist av,silist ac)
{
  int r = rand();
  int avlen = si_len(av);
  int sign;

  r = r % (avlen * 2);
  sign = r >= avlen ? 1 : -1;
  r = r >= avlen ? r - avlen : r;
  while(r--)
    av = av->next;
  return sign*av->next->key;
}



/*********************************************************** 
 * varibale with most often occurence with most often sign
 * in binary clauses, if several then 
 * variable with most often occurence, with most often sign 
 * Idea of heuristic from eisele 
 * (Andreas Eisele <andreas@adler.ims.uni-stuttgart.de>)
 **********************************************************/

int heuristic5(silist av,silist ac) 
{
  silist iav = av->next;
  silist iac;
  silist maxav = av;
  int* varinclauses;
  int no_bin_pos_occs = 0;
  int no_bin_neg_occs = 0;
  int no_pos_occs = 0;
  int no_neg_occs = 0;
  int max_no_bin_occs = 0;
  int max_no_occs = 0;
  int max_sign = 0;
  
  do {
    varinclauses = lit_to_clauses[iav->key];
    iac = ac->next;
    no_pos_occs = no_neg_occs = 0;
    no_bin_pos_occs = no_bin_neg_occs = 0;
    for(;;) /* ugly endless loop for speed, direct abort if end condition */
      {
	if(GETtaclauseid(iac->key) < abs(*varinclauses))
	  {
	    iac = iac->next; if (iac->next == iac) break;
	  }
	else if(GETtaclauseid(iac->key) == abs(*varinclauses))
	  {
	    if (*varinclauses > 0) 
	      {	
		no_pos_occs++;
		if (GETtaclauselen(iac->key) == 2)
		  no_bin_pos_occs++;
	      }
	    else
	      {
		no_neg_occs++;
		if (GETtaclauselen(iac->key) == 2)
		  no_bin_neg_occs++;
	      }
	    iac = iac->next; if (iac->next == iac)  break;
	    varinclauses++; if (*varinclauses == 0) break;
	  }  
	else      
	  { varinclauses++; if (*varinclauses == 0) break; }
      } /* while (*varinclauses && (iac->next != iac)) */
    if ((no_bin_pos_occs+no_bin_neg_occs) > max_no_bin_occs) 
      { 
	maxav = iav; 
	max_no_bin_occs = no_bin_pos_occs+no_bin_neg_occs;
	max_no_occs = no_pos_occs + no_neg_occs;
	max_sign = no_bin_pos_occs+no_pos_occs >  no_bin_neg_occs+no_neg_occs ? 1 : -1;
      }
    else if ((no_bin_pos_occs+no_bin_neg_occs) == max_no_bin_occs) 
      {
	if ((no_pos_occs + no_neg_occs) > max_no_occs)
	  {
	    maxav = iav; 
	    max_no_occs = no_pos_occs + no_neg_occs;
	    max_sign = no_bin_pos_occs+no_pos_occs >  no_bin_neg_occs+no_neg_occs ? 1 : -1;
	  }
      }
    iav = iav->next;
  }  while(iav->next != iav);
  return max_sign*maxav->key;
}



/*********************************************************** 
 * literal X with max. vector under lexicographic order
 *      (H_2(X),H_3(X),....)
 * where H_i(X) is
 *      1*max(h_i(X),h_i(-X)) + 2*min(h_i(X),h_i(-X))
 * where h_i(X) is the number of clauses of length i containing X 
 * Idea of heuristic from M. Boehm : 
 * boehm@engels1.cs.uni-duesseldorf.de
 **********************************************************/

int H_compare(int* p,int *n,int *mp,int *mn)
{
  int i=2;
  int diff;

  while(i < 100)
    {
      diff = max(p[i],n[i]) + 2*min(p[i],n[i]) - max(mp[i],mn[i]) + 2*min(mp[i],mn[i]);
      if (diff > 0)
	return 1;
      else if (diff < 0)
	return 0;
      i++;
    }
  return 0;
}

int heuristic6(silist av,silist ac) 
{
  silist iav = av->next;
  silist iac;
  silist maxav = av;
  int* varinclauses;
  int i;
  int no_pos_occs[100];
  int no_neg_occs[100];
  int max_no_pos_occs[100];
  int max_no_neg_occs[100];
  int max_sign;

  for (i=2;i < 100;i++)
    max_no_pos_occs[i] = max_no_neg_occs[i] = 0;
  do {
    varinclauses = lit_to_clauses[iav->key];
    iac = ac->next;
    for (i=2;i < 100;i++)
      no_pos_occs[i] = no_neg_occs[i] = 0;
    for(;;) /* ugly endless loop for speed, direct abort if end condition */
      {
	if(GETtaclauseid(iac->key) < abs(*varinclauses))
	  {
	    iac = iac->next; if (iac->next == iac) break;
	  }
	else if(GETtaclauseid(iac->key) == abs(*varinclauses))
	  {
	    if (*varinclauses > 0) 
	      {	
		if (GETtaclauselen(iac->key) < 100)
		  no_pos_occs[GETtaclauselen(iac->key)]++;
	      }
	    else
	      {
		if (GETtaclauselen(iac->key) < 100)
		  no_pos_occs[GETtaclauselen(iac->key)]++;
	      }
	    iac = iac->next; if (iac->next == iac)  break;
	    varinclauses++; if (*varinclauses == 0) break;
	  }  
	else      
	  { varinclauses++; if (*varinclauses == 0) break; }
      } /* while (*varinclauses && (iac->next != iac)) */
    if (H_compare(no_pos_occs,no_neg_occs,max_no_pos_occs,max_no_neg_occs)) 
      { 
	maxav = iav; 
	for (i=2;i<100;i++)
	  {
	    max_no_neg_occs[i] = no_neg_occs[i];
	    max_no_pos_occs[i] = no_pos_occs[i];
	  }
	max_sign = 1;
      }
    else if (H_compare(no_neg_occs,no_pos_occs,max_no_neg_occs,max_no_pos_occs)) 
      {
	maxav = iav; 
	for (i=2;i<100;i++)
	  {
	    max_no_neg_occs[i] = no_neg_occs[i];
	    max_no_pos_occs[i] = no_pos_occs[i];
	  }
	max_sign = -1;
      }
    iav = iav->next;
  }  while(iav->next != iav);
  return max_sign*maxav->key;
}



/* which one to choose */
int choose_variable(silist av,silist ac)
{
#ifdef CHECK
  if (av->next->next == av->next)
    error("should not be");
#endif
  switch(heuristic) {
  case 0  : return heuristic0(av,ac);
            break;
  case 1  : return heuristic1(av,ac);
            break;
  case 2  : return heuristic2(av,ac);
            break;
  case 3  : return heuristic3(av,ac);
            break;
  case 4  : return heuristic4(av,ac);
            break;
  case 5  : return heuristic5(av,ac);
            break;
  case 6  : return heuristic6(av,ac);
            break;
  default : error("unknown heuristic");exit(-1);
  } 
}


